#ifndef __STATE_H__
#define __STATE_H__

#include "driver/spi_master.h"
#include "driver/gpio.h"
#include "debounce.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "configuration.h"

#define MAX_PROCESS     (TOT_OUTPUTS*8*2)

#define ASYNC_TASK_PRIORITY     5
#define ADC_EVENT_BITS          0x01
#define INPUT_EVENT_BITS          0x02

#define ADD_REV_HIGH  0x00
#define ADD_REV_LOW   0x01
#define ADD_CONF_HIGH 0x00
#define ADD_CONF_LOW  0x20

#define WAIT4SEM      (20/portTICK_PERIOD_MS)

typedef struct {
    uint8_t localoutputs;
    uint8_t localinputs;
    uint8_t rawlocalinputs;
    uint16_t localpress[NUM_PRESS];
    uint16_t localtemp[NUM_TEMP];
    uint16_t local420[NUM_420];
    uint16_t localadc[NUM_ADC];
    uint8_t localdac[NUM_DAC];
    uint8_t externoutputs[NUM_EXT_OUTPUTS*2];
    uint8_t externinputs[TOT_EXT_INPUTS];
    uint8_t rawexterninputs[TOT_EXT_INPUTS];
    uint16_t externpress[NUM_EXT_PRESS*2];
    uint16_t externtemp[NUM_EXT_TEMP*2];
    uint16_t extern420[NUM_EXT_420*2];
    uint16_t externadc[NUM_EXT_ADC*2];
    uint8_t externdac[NUM_EXT_DAC*2];
    uint8_t activitybitmap[TOT_OUTPUTS];
    uint8_t activitydac;
    colori_led_t colore[NUM_SLAVES];
    uint8_t reset;
    uint8_t outputs_before_reset[TOT_OUTPUTS];
} machine_state_t;
typedef enum {
    OUTPUT_SPI,
    DAC_SPI
} spi_enum_t;

typedef struct {
    spi_device_handle_t handle;
    spi_enum_t spidevice;
    uint8_t arg1;
    uint8_t data[16];
} spi_transaction_msg_t;

typedef enum {
    STATE_SET_OUTPUTS,
    STATE_CLEAR_OUTPUTS,
    STATE_UPDATE_OUTPUTS,
    STATE_SET_ACTIVITY_OUTPUTS,
    STATE_CLEAR_ACTIVITY_OUTPUTS,
    STATE_SET_ACTIVITY_DAC,
    STATE_CLEAR_ACTIVITY_DAC,
    STATE_UPDATE_DAC,
} internal_state_message_type_t;

typedef struct {
    internal_state_message_type_t type;
    int num;
    uint8_t bitmap[TOT_OUTPUTS];
    uint16_t value;
    int notvalid;
} internal_state_message_t;

void task_main_cycle(void *args);
uint8_t get_local_inputs(digin_filter_t *filter);

int queue_adc_value(uint16_t *buffer, uint16_t value, int max, int index);
unsigned int average_adc_value(uint16_t *buffer, int max);
void take_offset_press(machine_configuration_t *config);

// State WRITE
void update_output_state(uint8_t *bitmap);
void update_input_state(uint8_t *bitmap, uint8_t *resilient, uint8_t *inverted);
void update_single_dac_state(int num, uint8_t value);
int update_dac_state(uint8_t *buffer);
void clear_output_state(uint8_t* bitmap);
void set_output_state(uint8_t *bitmap);
void update_output_state(uint8_t *bitmap);
void set_activity_bitmap_output(uint8_t *bitmap);
void set_activity_bitmap_dac(uint8_t bitmap);
void clear_activity_bitmap_output(uint8_t *bitmap);
void clear_activity_bitmap_dac(uint8_t bitmap);
void set_color_state(uint8_t *buffer);

// State READ
int pack_slave_state(uint8_t *buffer, uint8_t pulsingled, int slave);
int unpack_slave_state(uint8_t *buffer, int slave);
void get_color_state(uint8_t *buffer);
void get_output_state(uint8_t *bitmap);
int check_activity_bitmap_dac(uint8_t buffer);
int check_activity_bitmap_output(uint8_t *buffer);
int get_dac_state(uint8_t *buffer);
int get_adc_state(uint8_t *buffer);
int get_420ma_state(uint8_t *buffer);
int get_pure420ma_state(uint16_t *buffer);
int get_press_state(uint8_t *buffer, uint16_t *offsets);
int get_temps_state(uint8_t *buffer);
int get_input_state(uint8_t *bitmap);
void get_output_state(uint8_t *bitmap);
uint8_t get_single_output_state(int num);
int pack_state(uint8_t *buffer, uint16_t *offsets);
void init_state_manager(machine_configuration_t *config);
void reset_resilient_inputs();
void init_digin(machine_configuration_t *config);
void takeStateSemaphore(); // TODO: togliere
void giveStateSemaphore();
void send_output_state(internal_state_message_type_t type, uint8_t *bitmap, int notvalid);
void send_dac_state(int value, int num, int notvalid);
void manage_message(internal_state_message_t message);
void send_clear_activity_dac(uint8_t dac, int notvalid);
int is_in_reset();
#endif