#include <string.h>
#include "hardwareprofile.h"
#include "state.h"
#include "asynctasks.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/timers.h"
#include "system.h"
#include "driver/uart.h"
#include "listx.h"
#include "pcb.h"
#include "utility.h"
#include "dac_spi.h"
#include "adc_spi.h"

#define TAG "Async tasks"
#define INTERVAL_CHECK_MS   20

#define DEPRESSED(self)     if (self->todie) kill_self(self);
//#define DEPRESSED(self)     //if (self->todie) vTaskDelay(1000/portTICK_PERIOD_MS);

myLIST_HEAD(process_list);
SemaphoreHandle_t async_process_sem;

void init_async_process_management() {
    async_process_sem = xSemaphoreCreateMutex(); 
    init_process_list();
}

void kill_self(pcb_t *self) {
    timer_args_t *params = self->args;
    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    out_procq(&process_list, self);
    free_pcb(self);
    free(params);
    xSemaphoreGive(async_process_sem);
    vTaskDelete(NULL);
}

void terminate_timer(pcb_t *self) {
    timer_args_t *params = self->args;
    uint8_t map;
    //clear_output_state(params->bitmap);
    send_output_state(STATE_CLEAR_OUTPUTS, params->bitmap, self->todie);
    //xQueueSend(params->queue, NULL, 0);

    //clear_activity_bitmap_output(params->bitmap);
    send_output_state(STATE_CLEAR_ACTIVITY_OUTPUTS, params->bitmap, self->todie);
    map = params->dac >= 0 ? 1 << params-> dac : 0;
    //clear_activity_bitmap_dac(map);
    send_clear_activity_dac(map, self->todie);

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    ESP_LOGI(TAG, "Finished scaled dac task after %d milliseconds", params->timer*100);
    //kill_self(self);
    out_procq(&process_list, self);
    free_pcb(self);
    free(params);
    xSemaphoreGive(async_process_sem);
    vTaskDelete(NULL);
}


int check_ma420_reached(pcb_t *aux, uint16_t limit420ma, uint16_t *ma420) {
    timer_args_t *params = aux->args;
    if (params->num420ma >= 0 && params->limit420ma > 0) {
        if ((params->direction == 1 && ma420[params->num420ma] <= limit420ma) ||
            (params->direction == 0 && ma420[params->num420ma] >= limit420ma)) {
            return 1;
        }
    }
    return 0;
}


void _delete_task(pcb_t *todel) {
    timer_args_t *ptr;
    uint8_t map;

    ESP_LOGI(TAG, "deleteting process %d (%d) with alarm %i",(uint32_t) todel->task, (uint32_t)xTaskGetCurrentTaskHandle(), todel->alarm);
    ptr = (timer_args_t*) todel->args;
    takeStateSemaphore();
    ESP_LOGI(TAG, "about to delete");
    vTaskDelete(todel->task);
    ESP_LOGI(TAG, "deleted");
    giveStateSemaphore();
    clear_activity_bitmap_output(ptr->bitmap);
    map = ptr->dac >= 0 ? 1 << ptr-> dac : 0;
    clear_activity_bitmap_dac(map);
    // Clear outputs
    clear_output_state(ptr->bitmap);
    //if (ptr->dac >= 0 && ptr->dac < TOT_DAC)
        //update_single_dac_state(ptr->dac, 0);

    free(todel->args);
    out_procq(&process_list, todel);
    free_pcb(todel);
    ESP_LOGI(TAG, "freed memory");
}

void __delete_task(pcb_t *todel) {
    timer_args_t *ptr;
    uint8_t map;

    // Imposta la cancellazione del processo
    todel->todie = 1;
    ESP_LOGI(TAG, "soft deleteting process %d with alarm %i",(uint32_t) todel, todel->alarm);
    ptr = (timer_args_t*) todel->args;
    clear_activity_bitmap_output(ptr->bitmap);
    map = ptr->dac >= 0 ? 1 << ptr-> dac : 0;
    clear_activity_bitmap_dac(map);
    // Clear outputs
    clear_output_state(ptr->bitmap);
}

void delete_all_tasks() {
    pcb_t *aux;
    
    xSemaphoreTake(async_process_sem, portMAX_DELAY);

    while((aux = remove_procq(&process_list)) != NULL) {
        ESP_LOGI(TAG, "deleteting process");
        _delete_task(aux);
    }
    xSemaphoreGive(async_process_sem);
    vTaskDelay(1);
}

void delete_tasks_by_output(uint8_t *bitmap) {
    pcb_t *aux;
    int i, index = 0;
    timer_args_t *args;
    pcb_t *todelete[MAX_PROCESS];
    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    list_for_each_entry(aux, &process_list, p_next) {
        for (i = 0; i < TOT_OUTPUTS; i++) {
            args = (timer_args_t*) aux->args;
            if ((bitmap[i] & args->bitmap[i]) != 0) {
                ESP_LOGI(TAG, "scheduling task for deletion: bitmap");
                hexprint(args->bitmap, TOT_OUTPUTS, TAG);
                todelete[index++] = aux;
                break;
            }
        }
    }
    for (i = 0; i < index; i++)
        _delete_task(todelete[i]);
    xSemaphoreGive(async_process_sem);
    if (index > 0)
       vTaskDelay(1);
}

void delete_tasks_by_alarm(uint8_t *bitmap, uint16_t *ma420) {
    pcb_t  *aux;
    pcb_t *todelete[MAX_PROCESS];
    int index, i, num, bit;

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    index = 0;
    list_for_each_entry(aux, &process_list, p_next) {
        if(aux->alarm >= 0) {
            num = aux->alarm/8;
            bit = aux->alarm%8;
            if (num < TOT_INPUTS && (bitmap[num] & 1 << bit) != 0) {
                ESP_LOGI(TAG, "scheduling process with alarm %i for deletion", aux->alarm);
                todelete[index++] = aux;
                continue;
            }
        }
    }
    for (i = 0; i < index; i++)
        _delete_task(todelete[i]);
    xSemaphoreGive(async_process_sem);
    if (index > 0)
        vTaskDelay(1);
}

void task_simple_simple_timer(void *args) {
    pcb_t *self = (pcb_t*) args;
    timer_args_t *params = self->args;
    
    ESP_LOGI(TAG, "Started simple simple timer task for %i milliseconds", params->timer*100);
    //set_output_state(params->bitmap);
    //xQueueSend(params->queue, NULL, 0);
    send_output_state(STATE_SET_OUTPUTS, params->bitmap, self->todie);
    vTaskDelay(params->timer*100/portTICK_PERIOD_MS);

    /*clear_output_state(params->bitmap);
    clear_activity_bitmap_output(params->bitmap);
    xQueueSend(params->queue, NULL, 0);

    ESP_LOGI(TAG, "Finished timer task after %d milliseconds", params->timer*100);
    kill_self(self);*/
    terminate_timer(self);
}

void task_simple_timer(void *args) {
    pcb_t *self = (pcb_t*) args;
    timer_args_t *params = self->args;
    uint8_t buffer[64];
    uint64_t finished;
    TickType_t delay;
    uint16_t ma420[TOT_420];
    char firstreached = 0;
    
    ESP_LOGI(TAG, "Started timer task %d for %i milliseconds, limits %i and %i", (uint32_t)self, params->timer*100, params->first420limit, params->limit420ma);
    if (params->f_dac)
        send_dac_state(params->dacstart, params->dac, self->todie);
        //update_single_dac_state(params->dac, params->dacstart);
    if (params->first420limit == 0)
        firstreached = 1;

    send_output_state(STATE_SET_OUTPUTS, params->bitmap, self->todie);
    //set_output_state(params->bitmap);
    //xQueueSend(params->queue, NULL, 0);
    finished = get_millis() + params->timer*100;

    while(params->timer == 0 || get_millis() < finished) {
        delay = params->timer == 0 ? portMAX_DELAY : (finished - get_millis())/portTICK_PERIOD_MS;
        xEventGroupWaitBits(params->eventgroup, ADC_EVENT_BITS, pdTRUE, pdFALSE, delay);
        //xSemaphoreTake(async_process_sem, portMAX_DELAY);
        DEPRESSED(self);
        get_pure420ma_state(ma420);
        get_420ma_state(buffer);
        ESP_LOGI(TAG, "value found: %i (%i) (%i %i)", ma420[params->num420ma], buffer[params->num420ma], params->first420limit, params->limit420ma);
        if (firstreached && check_ma420_reached(self, params->limit420ma, ma420)) {
            ESP_LOGI(TAG, "Second limit reached at %llu", esp_timer_get_time());
            break;
        }
        else if (!firstreached && params->f_dac && check_ma420_reached(self, params->first420limit, ma420)) {
            //update_single_dac_state(params->dac, params->dacfinish);
            send_dac_state(params->dacfinish, params->dac, self->todie);
            ESP_LOGI(TAG, "First limit reached!");
            params->direction =  ma420[params->num420ma] > params->limit420ma;
            firstreached = 1;
        }
        //xSemaphoreGive(async_process_sem);
    }
   /* xSemaphoreTake(async_process_sem, portMAX_DELAY);

    ESP_LOGI(TAG, "output cleared");

    //takeStateSemaphore();
    //if (!self->todie) {
        clear_output_state(params->bitmap);
        clear_activity_bitmap_output(params->bitmap);
        update_single_dac_state(params->dac, params->dacstart);
    //}
    //giveStateSemaphore();
    DEPRESSED(self)

    ESP_LOGI(TAG, "sending update %d", (uint32_t) self->task);
    xQueueSend(params->queue, NULL, 0);

    ESP_LOGI(TAG, "Finished timer task %d after %d milliseconds", (uint32_t)self->task, params->timer*100);
    xSemaphoreGive(async_process_sem);
    kill_self(self);*/
    terminate_timer(self);
}
void launch_simple_simple_timer_task(uint16_t timer, int alarm, uint8_t *bitmap, QueueHandle_t queue, EventGroupHandle_t group) {
    timer_args_t *args = (timer_args_t*) malloc(sizeof(timer_args_t));
    pcb_t *process;

    memset(args, 0, sizeof(timer_args_t));
    memcpy(args->bitmap, bitmap,  5);
    args->queue = queue;
    args->timer = timer;
    args->eventgroup = group;

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    set_activity_bitmap_output(bitmap); 
    process = alloc_pcb((void*)args);
    process->timestamp = get_millis();
    process->alarm = alarm;
    insert_procq(&process_list, process);
    xTaskCreate(task_simple_simple_timer, "simple simple timer", 4096, (void*)process, ASYNC_TASK_PRIORITY, &process->task);
    ESP_LOGI(TAG, "created timer process (alarm %i)", process->alarm);
    xSemaphoreGive(async_process_sem);
}

void launch_simple_timer_task(uint16_t timer, int alarm, uint8_t limit420ma, int8_t num420ma, int direction, uint8_t *bitmap, QueueHandle_t queue, EventGroupHandle_t group) {
    timer_args_t *args = (timer_args_t*) malloc(sizeof(timer_args_t));
    pcb_t *process;

    memset(args, 0, sizeof(timer_args_t));
    memcpy(args->bitmap, bitmap,  5);
    args->queue = queue;
    args->timer = timer;
    args->dac = -1;
    args->f_dac = 0;
    args->limit420ma = ma420_to_adc(limit420ma);
    args->first420limit = 0;
    args->num420ma = (int)num420ma;
    args->direction = direction;
    args->eventgroup = group;

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    set_activity_bitmap_output(bitmap); 
    process = alloc_pcb((void*)args);
    process->timestamp = get_millis();
    process->alarm = alarm;
    insert_procq(&process_list, process);
    xTaskCreate(task_simple_timer, "simple timer", 4096, (void*)process, ASYNC_TASK_PRIORITY, &process->task);
    ESP_LOGI(TAG, "created timer process (alarm %i)", process->alarm);
    xSemaphoreGive(async_process_sem);
}

void launch_timer_task(uint16_t timer, int alarm, uint8_t limit420ma, int8_t num420ma, int direction, uint8_t numdac, 
    uint8_t dacstart, uint8_t dacfinish, uint8_t first420limit, uint8_t *bitmap, QueueHandle_t queue, EventGroupHandle_t group) {
    timer_args_t *args = (timer_args_t*) malloc(sizeof(timer_args_t));
    pcb_t *process;

    memset(args, 0, sizeof(timer_args_t));
    memcpy(args->bitmap, bitmap,  5);
    args->queue = queue;
    args->timer = timer;
    args->dac = numdac;
    args->f_dac = 1;
    args->dacstart = dacstart;
    args->dacfinish = dacfinish;
    args->first420limit = ma420_to_adc(first420limit);
    args->limit420ma = ma420_to_adc(limit420ma);
    args->num420ma = (int)num420ma;
    args->direction = direction;
    args->eventgroup = group;

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    set_activity_bitmap_output(bitmap); 
    process = alloc_pcb((void*)args);
    process->timestamp = get_millis();
    process->alarm = alarm;
    insert_procq(&process_list, process);
    xTaskCreate(task_simple_timer, "timer", 4096, (void*)process, ASYNC_TASK_PRIORITY, &process->task);
    ESP_LOGI(TAG, "created timer process with ma420 limit %i (%i)", args->limit420ma, limit420ma);
    xSemaphoreGive(async_process_sem);
}

void task_pwm_timer(void *args) {
    pcb_t *self = (pcb_t*) args;
    timer_args_t *params = self->args;
    int i;
    
    ESP_LOGI(TAG, "Started pwm task for %i milliseconds", params->timer*100);

    for (i = 0; i < params->impulsi; i++) {
        set_output_state(params->bitmap);
        xQueueSend(params->queue, NULL, 0);
        vTaskDelay((params->timehigh*100)/portTICK_PERIOD_MS);
        DEPRESSED(self);

        clear_output_state(params->bitmap);
        xQueueSend(params->queue, NULL, 0);
        vTaskDelay((params->timelow*100)/portTICK_PERIOD_MS);
        DEPRESSED(self);
    }

   /* clear_activity_bitmap_output(params->bitmap);

    ESP_LOGI(TAG, "Finished pwm task after %d milliseconds", params->timer*100);
    kill_self(self);*/
    terminate_timer(self);
}

void launch_pwm_timer_task(uint16_t high, uint16_t low, int impulsi, uint8_t *bitmap, QueueHandle_t queue) {
    timer_args_t *args = (timer_args_t*) malloc(sizeof(timer_args_t));
    pcb_t *process;

    memset(args, 0, sizeof(timer_args_t));
    memcpy(args->bitmap, bitmap,  5);
    args->queue = queue;
    args->timer = (high+low)*impulsi;
    args->timehigh = high;
    args->timelow = low;
    args->impulsi = impulsi;
    args->dac = -1;
    args->limit420ma = -1;
    args->num420ma = -1;

    xSemaphoreTake(async_process_sem, portMAX_DELAY);

    set_activity_bitmap_output(bitmap); 
    process = alloc_pcb((void*)args);
    process->timestamp = get_millis();
    process->alarm = -1;
    insert_procq(&process_list, process);
    xTaskCreate(task_pwm_timer, "pwm timer", 4096, (void*)process, ASYNC_TASK_PRIORITY, &process->task);
    ESP_LOGI(TAG, "created process");
    xSemaphoreGive(async_process_sem);
}


void task_scaled_dac_timer(void *args) {
    pcb_t *self = (pcb_t*) args;
    timer_args_t *params = self->args;
    int i;
    uint8_t stepsramp1, stepsramp2;
    uint16_t dacvalue, timeleft1, timeleft2, ma420[TOT_420];
    TickType_t lastwaketime;
    uint64_t finished;

    stepsramp1 = (params->vdf1 - params->vdi1)/params->scaletime1;
    stepsramp2 = (params->vdf2 - params->vdi2)/params->scaletime2;
    timeleft1 = params->time1 - params->scaletime1;
    timeleft2 = params->time2 - params->scaletime2;
    
    ESP_LOGI(TAG, "Started scaled dac task for %i milliseconds, from %d to %d in %d", params->timer*100, params->vdi1, params->vdf1, params->scaletime1);
    ESP_LOGI(TAG, "Scale steps: %d", stepsramp2);

    //set_output_state(params->bitmap);
    send_output_state(STATE_SET_OUTPUTS, params->bitmap, self->todie);

    dacvalue = params->vdi1;
    ESP_LOGI(TAG, "Starting from %d", dacvalue);
    //update_single_dac_state(params->dac, dac_to_volts(dacvalue));
    send_dac_state(dac_to_volts(dacvalue), params->dac, self->todie);
    lastwaketime = xTaskGetTickCount();
    //xQueueSend(params->queue, NULL, 0);

    for (i = 0; i < params->scaletime1; i++) {
        dacvalue += stepsramp1;
        ESP_LOGI(TAG, "Moving to %d", dacvalue);
        send_dac_state(dac_to_volts(dacvalue), params->dac, self->todie);
        //update_single_dac_state(params->dac, dac_to_volts(dacvalue));
        //xQueueSend(params->queue, NULL, 80/portTICK_PERIOD_MS);
        vTaskDelayUntil(&lastwaketime, 100/portTICK_PERIOD_MS);
        DEPRESSED(self);
        get_pure420ma_state(ma420);
        if (check_ma420_reached(self, params->limit420ma, ma420))
            terminate_timer(self); // Finisce anche la funzione
    }

    finished = get_millis() + timeleft1*100;
    while(get_millis() < finished) {
        vTaskDelay(INTERVAL_CHECK_MS/portTICK_PERIOD_MS);
        DEPRESSED(self);
        get_pure420ma_state(ma420);
        if (check_ma420_reached(self, params->limit420ma, ma420))
            terminate_timer(self); // Finisce anche la funzione
    }

    ESP_LOGI(TAG, "Second half from %d to %d in %d", params->vdi2, params->vdf2, params->scaletime2);
    dacvalue = params->vdi2;
    ESP_LOGI(TAG, "Starting from %d", dacvalue);
    send_dac_state(dac_to_volts(dacvalue), params->dac, self->todie);
    //update_single_dac_state(params->dac, dac_to_volts(dacvalue));
    //xQueueSend(params->queue, NULL, 0);

    for (i = 0; i < params->scaletime2; i++) {
        dacvalue += stepsramp2;
        send_dac_state(dac_to_volts(dacvalue), params->dac, self->todie);
        //update_single_dac_state(params->dac, dac_to_volts(dacvalue));
        //xQueueSend(params->queue, NULL, 80/portTICK_PERIOD_MS);
        vTaskDelayUntil(&lastwaketime, 100/portTICK_PERIOD_MS);
        DEPRESSED(self);
        get_pure420ma_state(ma420);
        if (check_ma420_reached(self, params->limit420ma, ma420))
            terminate_timer(self); // Finisce anche la funzione
    }
    ESP_LOGI(TAG, "waiting end for %d ms", timeleft2*100);
    finished = get_millis() + timeleft2 * 100;
    while (get_millis() < finished) {
        vTaskDelay(INTERVAL_CHECK_MS / portTICK_PERIOD_MS);
        DEPRESSED(self);
        get_pure420ma_state(ma420);
        if (check_ma420_reached(self, params->limit420ma, ma420))
            terminate_timer(self);     // Finisce anche la funzione
    }

    terminate_timer(self);
}

void launch_scaled_dac_timer_task(uint16_t timer1, uint16_t timer2, int alarm, uint8_t dac, 
        uint8_t vdi1, uint8_t vdf1, uint8_t vdi2, uint8_t vdf2, uint8_t scaletime1, uint8_t scaletime2,
        uint8_t limit420ma, int8_t num420ma, int direction, uint8_t *bitmap, QueueHandle_t queue) {
    timer_args_t *args = (timer_args_t*) malloc(sizeof(timer_args_t));
    pcb_t *process;
    uint8_t dacbitmap = 1 << dac;

    memset(args, 0, sizeof(timer_args_t));
    memcpy(args->bitmap, bitmap,  5);
    args->queue = queue;
    args->timer = timer1;
    args->time1 = timer2;
    args->time2 = timer1-timer2;
    args->scaletime1 = scaletime1;
    args->scaletime2 = scaletime2;
    args->vdi1 = volts_to_dac(vdi1);
    args->vdi2 = volts_to_dac(vdi2);
    args->vdf1 = volts_to_dac(vdf1);
    args->vdf2 = volts_to_dac(vdf2);
    args->dac = dac;
    args->limit420ma = ma420_to_adc(limit420ma);
    args->num420ma = (int)num420ma;
    args->direction = direction;

    set_activity_bitmap_output(bitmap);
    set_activity_bitmap_dac(dacbitmap);
    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    process = alloc_pcb((void*)args);
    process->timestamp = get_millis();
    process->alarm = alarm;
    insert_procq(&process_list, process);
    xTaskCreate(task_scaled_dac_timer, "scaled dac timer", 4096, (void*)process, ASYNC_TASK_PRIORITY, &process->task);
    ESP_LOGI(TAG, "created scaled dac timer process");
    xSemaphoreGive(async_process_sem);
}


void get_output_timers(int num, uint16_t *timers) {
    int i;
    pcb_t *aux;
    timer_args_t *args;
    unsigned long time = get_millis();

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    for (i = 0; i < 8; i++) {
        timers[i] = 0;
        list_for_each_entry(aux, &process_list, p_next) {
            args = (timer_args_t*) aux->args; 
            if (args->bitmap[num] & (1<<i) && args->timer != 0) {
                if (args->timer == 0)
                    timers[i] = 0xFFFF;
                else
                    timers[i] = args->timer - (time-aux->timestamp)/100;
                break;
            }
        }
    }
    xSemaphoreGive(async_process_sem);
}


void task_timer_after_ma420(void *args) {
    pcb_t *self = (pcb_t*) args;
    timer_args_t *params = self->args;
    uint64_t finished, delay;
    uint16_t ma420[TOT_420];
    uint8_t inputs[TOT_INPUTS];
    EventBits_t bits = ADC_EVENT_BITS;

    char f_reached = 0;

    ESP_LOGI(TAG, "Started timer task (subordinated to ma420) for %i milliseconds", params->timer*100);
    finished = get_millis() + params->timer*100;

    // Un tempo nullo viene interpretato come operazione senza fine (assumendo di essere interrotto da fuori)
    while (params->timer == 0 || get_millis() < finished) {
        delay = params->timer == 0 ? portMAX_DELAY : (finished - get_millis())/portTICK_PERIOD_MS;
        //vTaskDelay(1);
        bits = xEventGroupWaitBits(params->eventgroup, ADC_EVENT_BITS, pdTRUE, pdFALSE, delay);
        DEPRESSED(self);
        if (bits | ADC_EVENT_BITS) {
            get_pure420ma_state(ma420);
            if (check_ma420_reached(self, params->first420limit, ma420)) {
                f_reached = 1;
                ESP_LOGI(TAG, "Primo limite raggiunto: %i (di %i) %i", params->first420limit, params->limit420ma, ma420[params->num420ma]);
                break;
            } else {
                ESP_LOGI(TAG, "valore: %i (%i).", ma420[params->num420ma], params->first420limit);
            }
        }
    }

    if (f_reached) {
        params->direction =  ma420[params->num420ma] > params->limit420ma;
        //set_output_state(params->bitmap);
        send_output_state(STATE_SET_OUTPUTS, params->bitmap, self->todie);

        while (params->timer == 0 || get_millis() < finished) {
            delay = params->timer == 0 ? portMAX_DELAY : (finished - get_millis()) / portTICK_PERIOD_MS;
            bits = xEventGroupWaitBits(params->eventgroup, ADC_EVENT_BITS | INPUT_EVENT_BITS, pdTRUE, pdFALSE, delay);
            DEPRESSED(self);
            if (bits | ADC_EVENT_BITS) {
                get_pure420ma_state(ma420);
                if (check_ma420_reached(self, params->limit420ma, ma420)) {
                    ESP_LOGI(TAG, "Secondo limite raggiunto: %i %i", params->limit420ma, ma420[params->num420ma]);
                    break;
                }
            }
            if (bits | INPUT_EVENT_BITS) {
                get_input_state(inputs);
                if (check_bitmap_intersection(inputs, params->inputbitmap, TOT_INPUTS)) {
                    ESP_LOGI(TAG, "ingresso finito");
                    break;
                }
            }
        }
        clear_output_state(params->bitmap);
    }
    /*//takeStateSemaphore();
    //if (!self->todie)
        clear_activity_bitmap_output(params->bitmap);
    //giveStateSemaphore();
    DEPRESSED(self);

    ESP_LOGI(TAG, "Finished supervisor task after %d milliseconds", params->timer*100);
    xQueueSend(params->queue, NULL, 0);

    kill_self(self);*/
    terminate_timer(self);
}

void launch_timer_after_ma420_task(uint16_t timer, uint8_t limit420ma, int8_t num420ma, int direction,
    uint8_t first420limit, uint8_t *inputbitmap, uint8_t *outputbitmap, QueueHandle_t queue, EventGroupHandle_t group) {
    timer_args_t *args = (timer_args_t*) malloc(sizeof(timer_args_t));
    pcb_t *process;

    memset(args, 0, sizeof(timer_args_t));
    memcpy(args->bitmap, outputbitmap,  TOT_OUTPUTS);
    memcpy(args->inputbitmap, inputbitmap,  TOT_INPUTS);
    args->queue = queue;
    args->eventgroup = group;
    args->timer = timer;
    args->f_dac = 0;
    args->first420limit = ma420_to_adc(first420limit);
    args->limit420ma = ma420_to_adc(limit420ma);
    args->num420ma = (int)num420ma;
    args->direction = direction;

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    set_activity_bitmap_output(outputbitmap); 
    process = alloc_pcb((void*)args);
    process->timestamp = get_millis();
    process->alarm = -1;
    insert_procq(&process_list, process);
    ESP_LOGI(TAG, "created timer process with ma420 limits %i, %i", args->first420limit, args->limit420ma);
    xTaskCreate(task_timer_after_ma420, "timer", 4096, (void*)process, ASYNC_TASK_PRIORITY+1, &process->task);
    xSemaphoreGive(async_process_sem);
}

void task_timer_after_input(void *args) {
    pcb_t *self = (pcb_t*) args;
    timer_args_t *params = self->args;
    uint64_t finished, delay;
    uint8_t inputs[TOT_INPUTS];
    uint8_t buffer[TOT_INPUTS];
    uint16_t ma420[TOT_420];
    EventBits_t bits;

    char f_reached = 0;

    ESP_LOGI(TAG, "Started timer task (subordinated to input) for %i milliseconds", params->timer*100);
    finished = get_millis() + params->timer*100;
    memset(buffer, 0, TOT_INPUTS);
    set_single_bitmap(params->starter, buffer);

    // Un tempo nullo viene interpretato come operazione senza fine (assumendo di essere interrotto da fuori)
    while (params->timer == 0 || get_millis() < finished) {
        delay = params->timer == 0 ? portMAX_DELAY : (finished - get_millis())/portTICK_PERIOD_MS;
        bits  = xEventGroupWaitBits(params->eventgroup, INPUT_EVENT_BITS, pdTRUE, pdFALSE, delay);
        DEPRESSED(self);
        if (bits | INPUT_EVENT_BITS) {
            get_input_state(inputs);
            if (check_bitmap_intersection(inputs, buffer, TOT_INPUTS)) {
                ESP_LOGI(TAG, "ingresso finito");
                f_reached = 1;
                break;
            }
        }
    }

    if (f_reached) {
        params->direction =  ma420[params->num420ma] > params->limit420ma;
        //set_output_state(params->bitmap);
        send_output_state(STATE_SET_OUTPUTS, params->bitmap, self->todie);

        while (params->timer == 0 || get_millis() < finished) {
            delay = params->timer == 0 ? portMAX_DELAY : (finished - get_millis()) / portTICK_PERIOD_MS;
            bits = xEventGroupWaitBits(params->eventgroup, ADC_EVENT_BITS | INPUT_EVENT_BITS, pdTRUE, pdFALSE, delay);
            DEPRESSED(self);
            if (bits | ADC_EVENT_BITS) {
                get_pure420ma_state(ma420);
                if (check_ma420_reached(self, params->limit420ma, ma420)) {
                    ESP_LOGI(TAG, "Secondo limite raggiunto: %i %i", params->limit420ma, ma420[params->num420ma]);
                    break;
                }
            }
            if (bits | INPUT_EVENT_BITS) {
                get_input_state(inputs);
                if (check_bitmap_intersection(inputs, params->inputbitmap, TOT_INPUTS)) {
                    ESP_LOGI(TAG, "ingresso finito");
                    break;
                }
            }
        }
        //clear_output_state(params->bitmap);
    }
    /*clear_activity_bitmap_output(params->bitmap);
    ESP_LOGI(TAG, "Finished supervisor task after %d milliseconds", params->timer*100);
    xQueueSend(params->queue, NULL, 0);

    kill_self(self);*/
    terminate_timer(self);
}

void launch_timer_after_input_task(uint16_t timer, int starter, uint8_t limit420ma, int8_t num420ma, int direction,
        uint8_t *inputbitmap, uint8_t *outputbitmap, QueueHandle_t queue, EventGroupHandle_t group) {
    timer_args_t *args = (timer_args_t*) malloc(sizeof(timer_args_t));
    pcb_t *process;

    memset(args, 0, sizeof(timer_args_t));
    memcpy(args->bitmap, outputbitmap,  TOT_OUTPUTS);
    memcpy(args->inputbitmap, inputbitmap,  TOT_INPUTS);
    args->queue = queue;
    args->eventgroup = group;
    args->timer = timer;
    args->f_dac = 0;
    args->limit420ma = ma420_to_adc(limit420ma);
    args->num420ma = (int)num420ma;
    args->direction = direction;
    args->starter = starter;

    xSemaphoreTake(async_process_sem, portMAX_DELAY);
    set_activity_bitmap_output(outputbitmap); 
    process = alloc_pcb((void*)args);
    process->timestamp = get_millis();
    process->alarm = -1;
    insert_procq(&process_list, process);
    xTaskCreate(task_timer_after_input, "timer", 4096, (void*)process, ASYNC_TASK_PRIORITY+1, &process->task);
    ESP_LOGI(TAG, "created timer process with ma420 limit %i (%i)", args->limit420ma, limit420ma);
    xSemaphoreGive(async_process_sem);
}