/******************************************************************************/
/*                                                                            */
/*  HSW snc - Casalecchio di Reno (BO) ITALY                                  */
/*  ----------------------------------------                                  */
/*                                                                            */
/*  modulo: 24XX16.c                                                          */
/*                                                                            */
/*      gestione EEPROM 24XX16                                                */
/*                                                                            */
/*  Autore: Mattia Maldini e Massimo Zanna                                    */
/*                                                                            */
/*  Data  : 02/01/2017      REV  : 00.0                                       */
/*                                                                            */
/*  U.mod.: 04/01/2017      REV  : 15.0                                       */
/*                                                                            */
/******************************************************************************/

#include "i2c_devices/24XX16.h"
#include "driver/i2c.h"
#include <stdint.h>

static gpio_num_t PIN_WP;

void init24XX16(gpio_num_t writeProtect) {
    gpio_config_t io_conf;

    PIN_WP = writeProtect;

    // disable interrupt
    io_conf.intr_type = GPIO_PIN_INTR_DISABLE;
    // set as output mode
    io_conf.mode = GPIO_MODE_OUTPUT;
    // bit mask of the pins that you want to set,e.g.GPIO18/19
    io_conf.pin_bit_mask = BITMAP(PIN_WP);
    // disable pull-down mode
    io_conf.pull_down_en = 0;
    // disable pull-up mode
    io_conf.pull_up_en = 0;
    // configure GPIO with the given settings
    gpio_config(&io_conf);
    gpio_set_level(PIN_WP, 1);
}

static void writeProtectDisable() { gpio_set_level(PIN_WP, 0); }

static void writeProtectEnable() { gpio_set_level(PIN_WP, 1); }

esp_err_t EEAckPolling(unsigned char control) {
    esp_err_t        ret;
    i2c_cmd_handle_t cmd;
    unsigned long    counter = 0;

    cmd = i2c_cmd_link_create();
    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, control, 1);
    i2c_master_stop(cmd);

    do {
        counter++;
        ret = i2c_master_cmd_begin(I2C_NUM_0, cmd, 10 / portTICK_RATE_MS);
        vTaskDelay(portTICK_PERIOD_MS);
    } while (ret == ESP_FAIL && counter < 100);

    i2c_cmd_link_delete(cmd);

    return counter < 100 ? ESP_OK : ret;
}

/*----------------------------------------------------------------------------*/
/*  byteWrite_24XX16                                                          */
/*----------------------------------------------------------------------------*/
esp_err_t byteWrite_24XX16(uint8_t ControlByte, uint8_t HighAdd, uint8_t LowAdd, uint8_t data) {
    esp_err_t        ret;
    i2c_cmd_handle_t cmd;

    ControlByte &= 0xF1;
    ControlByte |= HighAdd << 1;

    if (HighAdd > HIGH_ADD_LIMIT)
        return ESP_FAIL;

    cmd = i2c_cmd_link_create();
    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, ControlByte, 1);
    i2c_master_write_byte(cmd, LowAdd, 1);
    i2c_master_write_byte(cmd, data, 1);
    i2c_master_stop(cmd);

    writeProtectDisable();
    ret = i2c_master_cmd_begin(I2C_NUM_0, cmd, 10 / portTICK_RATE_MS);
    writeProtectEnable();
    i2c_cmd_link_delete(cmd);

    if (ret != ESP_OK)
        return ret;

    ret = EEAckPolling(ControlByte);

    return ret;
}

esp_err_t byteRead_24XX16(uint8_t ControlByte, uint8_t HighAdd, uint8_t LowAdd, uint8_t *Data) {
    esp_err_t        ret;
    i2c_cmd_handle_t cmd;

    if (HighAdd > HIGH_ADD_LIMIT)
        return ESP_FAIL;

    ControlByte &= 0xF1;
    ControlByte |= HighAdd << 1;

    cmd = i2c_cmd_link_create();
    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, ControlByte, 1);
    i2c_master_write_byte(cmd, LowAdd, 1);

    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, ControlByte | 0x01, 1);
    i2c_master_read_byte(cmd, Data, I2C_MASTER_NACK);
    i2c_master_stop(cmd);

    ret = i2c_master_cmd_begin(I2C_NUM_0, cmd, 10 / portTICK_RATE_MS);

    i2c_cmd_link_delete(cmd);

    return ret;
}

esp_err_t pageWrite_24XX16(uint8_t ControlByte, uint8_t HighAdd, uint8_t LowAdd, uint8_t *wrptr, int Length) {
    esp_err_t        ret;
    i2c_cmd_handle_t cmd;

    ControlByte &= 0xF1;
    ControlByte |= HighAdd << 1;

    cmd = i2c_cmd_link_create();
    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, ControlByte, 1);
    i2c_master_write_byte(cmd, LowAdd, 1);
    i2c_master_write(cmd, wrptr, Length, 1);
    i2c_master_stop(cmd);

    writeProtectDisable();
    ret = i2c_master_cmd_begin(I2C_NUM_0, cmd, 10 / portTICK_PERIOD_MS);
    writeProtectEnable();
    i2c_cmd_link_delete(cmd);

    if (ret != ESP_OK)
        return ret;

    ret = EEAckPolling(ControlByte);     // perform Ack Polling

    return ret;
}

esp_err_t sequentialWrite_24XX16(uint8_t ControlByte, uint8_t HighAdd, uint8_t LowAdd, uint8_t *wrptr, uint Length) {
    esp_err_t ret;
    uint      page_size = 0;

    if (HighAdd > HIGH_ADD_LIMIT) {
        return ESP_FAIL;
    }

    ControlByte &= 0xF1;
    ControlByte |= HighAdd << 1;

    uint fullAdd = (HighAdd << 8) | LowAdd;

    if (fullAdd + Length >= EEPROM_SIZE) {
        return ESP_FAIL;
    }

    while (Length > 0) {
        page_size = PAGE_SIZE - (LowAdd % PAGE_SIZE);
        page_size = (Length < page_size) ? Length : page_size;

        ret = pageWrite_24XX16(ControlByte, HighAdd, LowAdd, wrptr, page_size);

        if (ret != ESP_OK)
            return ret;

        Length -= page_size;
        wrptr += page_size;
        // Adjust addresses
        if (LowAdd + page_size > 0xFF) {
            if (HighAdd == HIGH_ADD_LIMIT) {
                return ESP_FAIL;     // Overflow - should not happen here but in the
                                     // previous control
            } else {
                HighAdd++;
                LowAdd = 0x00;
            }
        } else {
            LowAdd += page_size;
        }
        ControlByte &= 0xF0;
        ControlByte &= 0xF1;
        ControlByte |= HighAdd << 1;
        ControlByte |= 0x00;     // R/W - write operation
    }
    return ESP_OK;
}

esp_err_t blockRead_24XX16(uint8_t ControlByte, uint8_t HighAdd, uint8_t LowAdd, uint8_t *rdptr, unsigned int length) {
    esp_err_t        ret;
    i2c_cmd_handle_t cmd;

    ControlByte &= 0xF1;
    ControlByte |= HighAdd << 1;

    cmd = i2c_cmd_link_create();
    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, ControlByte, 1);
    i2c_master_write_byte(cmd, LowAdd, 1);

    i2c_master_start(cmd);
    i2c_master_write_byte(cmd, ControlByte | 0x01, 1);
    i2c_master_read(cmd, rdptr, length, I2C_MASTER_LAST_NACK);
    i2c_master_stop(cmd);

    ret = i2c_master_cmd_begin(I2C_NUM_0, cmd, 10 / portTICK_RATE_MS);
    i2c_cmd_link_delete(cmd);

    return ret;
}

esp_err_t sequentialRead_24XX16(uint8_t ControlByte, uint8_t HighAdd, uint8_t LowAdd, uint8_t *rdptr, uint length) {
    // TODO check for page/block/device overlapping
    uint      block_size;
    esp_err_t ret;

    if (HighAdd > HIGH_ADD_LIMIT) {
        return ESP_FAIL;
    }

    ControlByte &= 0xF1;
    ControlByte |= HighAdd << 1;

    uint16_t fullAdd = (HighAdd << 8) | LowAdd;

    if (fullAdd + length >= EEPROM_SIZE) {
        return ESP_FAIL;
    }

    while (length > 0) {
        block_size = length;
        ret        = blockRead_24XX16(ControlByte, HighAdd, LowAdd, rdptr, block_size);

        if (ret != ESP_OK)
            return ret;

        rdptr += block_size;
        length -= block_size;

        if (length == 0) {
            break;
        }

        HighAdd++;
        LowAdd  = 0x00;
        fullAdd = 0x0000;

        ControlByte &= 0xF0;
        ControlByte &= 0xF1;
        ControlByte |= HighAdd << 1;
        ControlByte |= 0x00;     // R/W - write operation
    }
    return ESP_OK;
}