#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/timers.h"
#include "driver/gpio.h"
#include "esp_wifi.h"
#include "hardwareprofile.h"
#include "system.h"
#include "driver/uart.h"
#include "uart_primary.h"
#include "dac_spi.h"
#include "utility.h"
#include "i2c_devices/i2c_driver.h"
#include "i2c_devices/24XX16.h"
#include "packet_manager.h"
#include "serial485.h"
#include "state.h"
#include "debounce.h"
#include "asynctasks.h"
#include "adc_spi.h"
#include "esp_task_wdt.h"


#define MAIN_TAG "Main"
#define ADD         0x0202

#define CHECK_LEN(p, l) \
    if (p.lenght - HEADER_LEN != l){ \
        send_nack(p.sender,ADD, p.command.cmd, WRONG_FORMAT); \
        break; \
    }

#define CHECK_RESET(p) \
    if (is_in_reset()) { \
        send_nack(p.sender,ADD, p.command.cmd, IMPOSSIBLE); \
        break; \
    } \


unsigned char revision[11] = {1, 4, 19, 0, 1, 0, 0, 0, 0, 0 ,0};
machine_configuration_t *config_to_save;


void task_idle() {
    char       val = 0;
    TickType_t last;
    last = xTaskGetTickCount();
    while (1) {
        //ESP_LOGI("IDLE CHECK", "xPortGetFreeHeapSize: %d", xPortGetFreeHeapSize());
        gpio_set_level(GPIO_NUM_15, val);
        val = val ? 0 : 1;
        vTaskDelayUntil(&last, 1000 / portTICK_PERIOD_MS);
    }
    vTaskDelete(NULL);
}

void task_main(void *null) {
    static QueueHandle_t       mainqueue, updatequeue;
    static EventGroupHandle_t  updategroup;
    uint8_t             buffer[1024], tmpbuffer[32], buffer2[TOT_INPUTS+TOT_OUTPUTS];
    uint8_t value = 0, arg1, arg2, map = 0, vdi1, vdi2, vdf1, vdf2, numdac, dacstart, dacfinish, partial420;
    packet_t            packet;
    uint16_t            timer1, timer2, timer3, timer4;
    machine_configuration_t             config;
    config_to_save = &config;
    void *args[3] = {(void*) &updatequeue, (void*) &config, (void*) &updategroup};
    int len, i, num, alarm, num1;
    TickType_t last;
    uint16_t timers[8];

    default_configuration(&config);
    esp_efuse_mac_get_default(&revision[5]);

    mainqueue = xQueueCreate(64, sizeof(packet_t));
    updatequeue = xQueueCreate(1, 0);
    updategroup = xEventGroupCreate();
    esp_wifi_set_mode(WIFI_MODE_NULL);

    init_gpio();
    init_i2c(I2C_SDA, I2C_SCL);
    init24XX16(EEPROM_WP);

    //byteRead_24XX16(EEPROM_ADD, 0x00, 0x00, &map);
    if (map != 0xAA && 0) {
        byteWrite_24XX16(EEPROM_ADD, 0x00, 0x00, 0xAA);

        byteRead_24XX16(EEPROM_ADD, 0x00, 0x00, &map);
        if (map != 0xAA) {
            last = xTaskGetTickCount();
            // RAM Error
            for(;;){
                gpio_set_level(GPIO_NUM_15, value);
                value = value ? 0 : 1;
                vTaskDelayUntil(&last, 300 / portTICK_PERIOD_MS);
            }
        } else {
            sequentialWrite_24XX16(EEPROM_ADD, ADD_REV_HIGH, ADD_REV_LOW, revision, sizeof(revision));
            sequentialWrite_24XX16(EEPROM_ADD, ADD_REV_HIGH, ADD_REV_LOW, (uint8_t*)&config, sizeof(machine_configuration_t));
        }
    }
    else {
        sequentialRead_24XX16(EEPROM_ADD, ADD_REV_HIGH, ADD_REV_LOW, revision, sizeof(revision));
        sequentialRead_24XX16(EEPROM_ADD, ADD_REV_HIGH, ADD_REV_LOW, (uint8_t*)&config, sizeof(machine_configuration_t));
    }

    //FIXME: se metto l'attesa qui poi la comunicazione con lo slave non funziona...
    //vTaskDelay(400/portTICK_PERIOD_MS);

    //init_low_voltage_save();
    init_uart_primary(&mainqueue);
    init_async_process_management();
    init_state_manager(&config);

    xTaskCreate(task_idle, "idle", 2048, NULL, 1, NULL);
    xTaskCreate(task_slave_communication, "485 slaves task", 8192, args, 5, NULL);
    xTaskCreate(task_main_cycle, "update state task", 8192, args, 2, NULL);

    // Attendi un po' prima di cominciare a rispondere
    vTaskDelay(200/portTICK_PERIOD_MS);
    xQueueReset(mainqueue);

    esp_task_wdt_add(NULL);
    //TODO: considerare la configurazione
    for (;;) {
        esp_task_wdt_reset();

        if (xQueueReceive(mainqueue, (void *)&packet, 10 / portTICK_PERIOD_MS)) {
            //ESP_LOGI(MAIN_TAG, "Received command 0x%0X", packet.command.cmd);
            switch (packet.command.cmd) {
                case SET_REV_SCHEDA:
                    CHECK_LEN(packet, 5);
                    memcpy(revision, packet.data, 5);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    sequentialWrite_24XX16(EEPROM_ADD, ADD_REV_HIGH, ADD_REV_LOW, (uint8_t *)revision, 11);
                    break;
                case GET_REV_SCHEDA:
                    CHECK_LEN(packet, 0);
                    send_packet(packet.sender, ADD, packet.command.cmd, revision, 11);
                    break;
                case SET_OUTPUTS:
                    CHECK_LEN(packet, 6);
                    if (packet.data[5] == 0) {
                        for (i = 0; i < 5; i++)
                            packet.data[i] = ~packet.data[i];
                    }
                    for (i = 0; i < TOT_OUTPUTS; i++)
                        buffer[i] = packet.data[TOT_OUTPUTS-i-1];

                    if (0 && !check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS)) {
                        send_nack(packet.sender, ADD, packet.command.cmd, IMPOSSIBLE);
                        break;
                    }

                    turn_on_led(config, buffer);
                    update_output_state(buffer);
                    xQueueSend(updatequeue, NULL, 0 / portTICK_PERIOD_MS);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;
                case GET_OUTPUTS:
                    CHECK_LEN(packet, 0);
                    get_output_state(buffer);
                    send_packet(packet.sender, ADD, packet.command.cmd,buffer, TOT_OUTPUTS);
                    break;

                case GET_TEMPS:
                    CHECK_LEN(packet, 0);
                    get_temps_state(buffer);
                    send_packet(packet.sender, ADD, packet.command.cmd, buffer, TOT_TEMP);
                    break;

                case GET_PRESS:
                    CHECK_LEN(packet, 0);
                    get_press_state(buffer, config.offsetpress);
                    send_packet(packet.sender, ADD, packet.command.cmd, buffer, TOT_PRESS);
                    break;

                case GET_ADC:
                    CHECK_LEN(packet, 0);
                    get_adc_state(buffer);
                    send_packet(packet.sender, ADD, packet.command.cmd, buffer, TOT_ADC);
                    break;

                case RESET_0:
                    CHECK_LEN(packet, 0);
                    delete_all_tasks();
                    memset(buffer, 0, TOT_DAC*sizeof(uint16_t));
                    update_dac_state(buffer);
                    memset(buffer, 0, TOT_OUTPUTS);
                    turn_on_led(config, buffer);
                    update_output_state(buffer);
                    xQueueSend(updatequeue, NULL, 0 / portTICK_PERIOD_MS);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case GET_INPUTS:
                    CHECK_LEN(packet, 0);
                    get_input_state(buffer);
                    send_packet(packet.sender, ADD, packet.command.cmd, buffer, 3);
                    reset_resilient_inputs();
                    break;

                case GET_CONFIG:
                    CHECK_LEN(packet, 0);
                    len = pack_configuration(config,buffer);
                    send_packet(packet.sender, ADD, packet.command.cmd, buffer, len);
                    break;

                case SET_CONFIG:
                    CHECK_LEN(packet, 19);
                    unpack_configuration(&config, packet.data); 
                    init_digin(&config);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    sequentialWrite_24XX16(EEPROM_ADD, ADD_REV_HIGH, ADD_REV_LOW, (uint8_t *)config_to_save,
                                   sizeof(machine_configuration_t));
                    break;

                case SET_SINGLE_DA:
                    CHECK_LEN(packet, 2);
                    num = packet.data[0];
                    value = packet.data[1];
                    if (1 || (config.dac | 1 << num)){
                        update_single_dac_state(num, value);
                        xQueueSend(updatequeue, NULL, 0 / portTICK_PERIOD_MS);
                        send_ack(packet.sender,ADD, packet.command.cmd);
                    } else {
                        send_nack(packet.sender,ADD, packet.command.cmd, WRONG_CRC); 
                    }
                    break;

                case SET_RESET_0:
                    CHECK_LEN(packet, 2);
                    memset(buffer, 0, TOT_INPUTS);
                    set_single_bitmap(packet.data[0], buffer);

                    if (0 && !check_bitmap_inclusion(buffer, config.inputs, TOT_INPUTS)) {
                        send_nack(packet.sender, ADD, packet.command.cmd, IMPOSSIBLE);
                        break;
                    }

                    config.alarm = packet.data[0];
                    config.reset_outputs = packet.data[1];
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    sequentialWrite_24XX16(EEPROM_ADD, ADD_REV_HIGH, ADD_REV_LOW, (uint8_t *)config_to_save,
                                   sizeof(machine_configuration_t));
                    break;

                case GET_RESET_0:
                    CHECK_LEN(packet, 0);
                    buffer[0] = config.alarm;
                    buffer[1] = config.reset_outputs;
                    send_packet(packet.sender, ADD, packet.command.cmd, buffer, 2);
                    break;
                case GET_STATE:
                    CHECK_LEN(packet, 0);
                    len = pack_state(buffer, config.offsetpress);
                    send_packet(packet.sender, ADD, packet.command.cmd, buffer, len);
                    reset_resilient_inputs();
                    break;

                case SET_OUTPUTn_TIME:
                    CHECK_LEN(packet, 3);
                    CHECK_RESET(packet);
                    num    = packet.data[0];
                    timer1 = packet.data[1] << 8 | packet.data[2];
                    memset(buffer, 0, TOT_OUTPUTS);
                    set_single_bitmap((int)num, buffer);
                    
                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                    if (check_activity_bitmap_output(buffer)) {
                        // Se l'uscita e' gia' usata annulla la funzione precedente
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //send_nack(packet.sender, ADD, packet.command.cmd, IMPOSSIBLE);
                        //break;
                    }

                    launch_simple_simple_timer_task(timer1, -1, buffer, updatequeue, updategroup);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTS_TIME:
                    CHECK_LEN(packet, TOT_OUTPUTS+2);
                    CHECK_RESET(packet);
                    timer1 = packet.data[TOT_OUTPUTS] << 8 | packet.data[TOT_OUTPUTS+1];
                    for (i = 0; i < TOT_OUTPUTS; i++)
                        buffer[i] = packet.data[TOT_OUTPUTS-1-i];

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //send_nack(packet.sender, ADD, packet.command.cmd, IMPOSSIBLE);
                        //break;
                    }

                    launch_simple_simple_timer_task(timer1, -1, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTn_TIME_420MA:
                    CHECK_LEN(packet, 5);
                    CHECK_RESET(packet);
                    num    = packet.data[0];
                    timer1 = packet.data[1] << 8 | packet.data[2];
                    arg1   = packet.data[3];
                    arg2   = packet.data[4];
                    memset(buffer, 0, TOT_OUTPUTS);
                    set_single_bitmap((int)num, buffer);

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    get_420ma_state(tmpbuffer);
                    launch_simple_timer_task(timer1, -1, arg2, arg1, tmpbuffer[arg1] > arg2, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTn_TIME_420MA_DAC_REACT:
                    CHECK_LEN(packet, 9);
                    CHECK_RESET(packet);
                    num    = packet.data[0];
                    timer1 = packet.data[1] << 8 | packet.data[2];
                    arg1   = packet.data[3];
                    arg2   = packet.data[4];
                    numdac = packet.data[5];
                    dacstart = packet.data[6];
                    dacfinish = packet.data[7];
                    partial420 = packet.data[8];
                    memset(buffer, 0, TOT_OUTPUTS);
                    set_single_bitmap((int)num, buffer);

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    get_420ma_state(tmpbuffer);
                    launch_timer_task(timer1, -1, arg2, arg1, tmpbuffer[arg1] > partial420, numdac, dacstart, dacfinish, partial420, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTn_TIME_AFTER_420MA:
                    CHECK_LEN(packet, 7);
                    CHECK_RESET(packet);
                    num    = packet.data[0];
                    timer1 = packet.data[1] << 8 | packet.data[2];
                    arg1   = packet.data[3];
                    arg2   = packet.data[4];
                    partial420 = packet.data[5];
                    // Un allarme fuori scala viene interpretato come niente allarmi
                    alarm = packet.data[6] >= TOT_INPUTS*8 ? -1 : packet.data[6];
                    memset(buffer, 0, TOT_OUTPUTS);
                    memset(buffer2, 0, TOT_INPUTS);
                    set_single_bitmap((int)num, buffer);
                    if (alarm >= 0)
                        set_single_bitmap((int)alarm, buffer2);

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    get_420ma_state(tmpbuffer);
                    launch_timer_after_ma420_task(timer1, arg2, arg1, tmpbuffer[arg1] > partial420, partial420, buffer2, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTS_TIME_AFTER_420MA:
                    CHECK_LEN(packet, 11);
                    CHECK_RESET(packet);
                    timer1 = packet.data[TOT_OUTPUTS] << 8 | packet.data[TOT_OUTPUTS+1];
                    arg1   = packet.data[TOT_OUTPUTS+2];
                    arg2   = packet.data[TOT_OUTPUTS+3];
                    partial420 = packet.data[TOT_OUTPUTS+4];
                    // Un allarme fuori scala viene interpretato come niente allarmi
                    alarm = packet.data[TOT_OUTPUTS+5] >= TOT_INPUTS*8 ? -1 : packet.data[TOT_OUTPUTS+5];
                    for (i = 0; i < TOT_OUTPUTS; i++)
                        buffer[i] = packet.data[TOT_OUTPUTS-1-i];
                    memset(buffer2, 0, TOT_INPUTS);
                    if (alarm >= 0)
                        set_single_bitmap((int)alarm, buffer2);

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    get_420ma_state(tmpbuffer);
                    launch_timer_after_ma420_task(timer1, arg2, arg1, tmpbuffer[arg1] > partial420, partial420, buffer2, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTn_TIME_AFTER_INPUT:
                    CHECK_LEN(packet, 7);
                    CHECK_RESET(packet);
                    num1 = packet.data[0];
                    timer1 = packet.data[1] << 8 | packet.data[2];
                    arg1   = packet.data[3];
                    arg2   = packet.data[4];
                    num = packet.data[5];
                    alarm = packet.data[6] >= TOT_INPUTS*8 ? -1 : packet.data[6];

                    memset(buffer, 0, TOT_OUTPUTS);
                    set_single_bitmap(num1, buffer);

                    memset(buffer2, 0, TOT_INPUTS);
                    if (alarm >= 0)
                        set_single_bitmap((int)alarm, buffer2);

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    get_420ma_state(tmpbuffer);
                    launch_timer_after_input_task(timer1, num, arg2, arg1, tmpbuffer[arg1] > arg2, buffer2, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTS_TIME_420MA:
                    CHECK_LEN(packet, TOT_OUTPUTS+4);
                    CHECK_RESET(packet);
                    timer1 = packet.data[TOT_OUTPUTS] << 8 | packet.data[TOT_OUTPUTS+1];
                    arg1 = packet.data[TOT_OUTPUTS+2];
                    arg2 = packet.data[TOT_OUTPUTS+3];
                    for (i = 0; i < TOT_OUTPUTS; i++)
                        buffer[i] = packet.data[TOT_OUTPUTS-1-i];

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    get_420ma_state(tmpbuffer);
                    launch_simple_timer_task(timer1, -1, arg2, arg1, tmpbuffer[arg1] > arg2, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTS_TIME_420MA_DAC_REACT:
                    CHECK_LEN(packet, TOT_OUTPUTS+8);
                    CHECK_RESET(packet);
                    timer1 = packet.data[TOT_OUTPUTS] << 8 | packet.data[TOT_OUTPUTS+1];
                    arg1 = packet.data[TOT_OUTPUTS+2];
                    arg2 = packet.data[TOT_OUTPUTS+3];
                    numdac = packet.data[TOT_OUTPUTS+4];
                    dacstart = packet.data[TOT_OUTPUTS+5];
                    dacfinish = packet.data[TOT_OUTPUTS+6];
                    partial420 = packet.data[TOT_OUTPUTS+7];
                    for (i = 0; i < TOT_OUTPUTS; i++)
                        buffer[i] = packet.data[TOT_OUTPUTS-1-i];

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    get_420ma_state(tmpbuffer);
                    launch_timer_task(timer1, -1, arg2, arg1, tmpbuffer[arg1] > partial420, numdac, dacstart, dacfinish, partial420, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case GET_OUTPUTS1_TIME ... GET_OUTPUTS5_TIME:
                    CHECK_LEN(packet, 0);
                    num = packet.command.parts.subcommand & 0x0F;

                    buffer[0] = get_single_output_state(num);
                    get_output_timers(num, timers);
                    for (i = 0; i < 8; i++) {
                        buffer[1+i*2] = timers[i] >> 8;
                        buffer[2+i*2] = timers[i] & 0xFF;
                    }

                    send_packet(packet.sender, ADD, packet.command.cmd, buffer, 17);
                    break;

                case SET_OUTPUTn_PWM:
                    CHECK_LEN(packet, 6);
                    CHECK_RESET(packet);
                    num = packet.data[0];
                    timer1 = packet.data[1] << 8 | packet.data[2];
                    timer2 = packet.data[3] << 8 | packet.data[4];
                    arg1 = packet.data[5];
                    memset(buffer, 0, TOT_OUTPUTS);
                    set_single_bitmap((int)num, buffer);

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    launch_pwm_timer_task(timer1, timer2, (int)arg1, buffer, updatequeue);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTS_PWM:
                    CHECK_LEN(packet, TOT_OUTPUTS+5);
                    CHECK_RESET(packet);
                    num = packet.data[0];
                    timer1 = packet.data[TOT_OUTPUTS] << 8 | packet.data[TOT_OUTPUTS+1];
                    timer2 = packet.data[TOT_OUTPUTS+2] << 8 | packet.data[TOT_OUTPUTS+3];
                    arg1 = packet.data[TOT_OUTPUTS+4];
                    for (i = 0; i < TOT_OUTPUTS; i++)
                        buffer[i] = packet.data[TOT_OUTPUTS-1-i];

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    launch_pwm_timer_task(timer1, timer2, (int)arg1, buffer, updatequeue);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTn_ALARM:
                    CHECK_LEN(packet, 4);
                    CHECK_RESET(packet);
                    num = packet.data[0];
                    timer1 = packet.data[1] << 8 | packet.data[2];
                    arg1 = packet.data[3];
                    memset(buffer, 0, TOT_OUTPUTS);
                    set_single_bitmap((int)num, buffer);

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    launch_simple_simple_timer_task(timer1, arg1, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTS_ALARM:
                    CHECK_LEN(packet, 8);
                    CHECK_RESET(packet);
                    timer1 = packet.data[TOT_OUTPUTS] << 8 | packet.data[TOT_OUTPUTS+1];
                    arg1 = packet.data[TOT_OUTPUTS+2];
                    for (i = 0; i < TOT_OUTPUTS; i++)
                        buffer[i] = packet.data[TOT_OUTPUTS-1-i];

                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //break;
                    }

                    launch_simple_simple_timer_task(timer1, arg1, buffer, updatequeue, updategroup);
                    send_ack(packet.sender,ADD, packet.command.cmd);
                    break;

                case GET_420MA:
                    CHECK_LEN(packet, 0);
                    get_420ma_state(buffer);
                    hexprint(buffer, TOT_420, MAIN_TAG);
                    send_packet(packet.sender, ADD, packet.command.cmd, buffer, TOT_420);
                    break;

                case RESET_PORT:
                    CHECK_LEN(packet, 1);
                    num = packet.data[0];
                    memset(buffer, 0, TOT_OUTPUTS);
                    buffer[num] = 0xFF;
                    delete_tasks_by_output(buffer);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;

                case RESET_PORT_BITMAP:
                    CHECK_LEN(packet, 2);
                    num = packet.data[0];
                    memset(buffer, 0, TOT_OUTPUTS);
                    buffer[num] = packet.data[1];
                    delete_tasks_by_output(buffer);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;

                case SET_MULTI_DAC:
                    CHECK_LEN(packet, TOT_DAC);
                    if(check_activity_bitmap_dac(0xFF)) {
                        send_nack(packet.sender, ADD, packet.command.cmd, IMPOSSIBLE);
                        break;
                    }
                    buffer[0] = packet.data[0];
                    buffer[1] = packet.data[1];
                    buffer[2] = packet.data[2];
                    buffer[3] = packet.data[3];
                    buffer[4] = packet.data[4];
                    buffer[5] = packet.data[5];
                    update_dac_state(buffer);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTn_SCALED_DAC:
                    CHECK_LEN(packet, 12);
                    CHECK_RESET(packet);
                    num = packet.data[0];
                    timer1 = packet.data[1] << 8 | packet.data[2];
                    numdac = packet.data[3];
                    vdi1 = packet.data[4];
                    vdf1 = packet.data[5];
                    timer3 = packet.data[6];
                    vdi2 = packet.data[7];
                    vdf2 = packet.data[8];
                    timer2 = packet.data[9] << 8 | packet.data[10];
                    timer4 = packet.data[11];
                    memset(buffer, 0, TOT_OUTPUTS);
                    set_single_bitmap((int)num, buffer);
                    map = 1 << numdac;
                    
                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer) ||
                            check_activity_bitmap_dac(map)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //ESP_LOGI(MAIN_TAG, "Not possible, resource is already allocated");
                        //send_nack(packet.sender, ADD, packet.command.cmd, IMPOSSIBLE);
                        //break;
                    }

                    launch_scaled_dac_timer_task(timer1, timer2, -1, numdac, vdi1, 
                            vdf1, vdi2, vdf2, timer3, timer4,
                            0, -1, 0, buffer, updatequeue);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTS_SCALED_DAC:
                    CHECK_LEN(packet, 16);
                    CHECK_RESET(packet);
                    timer1 = packet.data[5] << 8 | packet.data[6];
                    numdac = packet.data[7];
                    vdi1 = packet.data[8];
                    vdf1 = packet.data[9];
                    timer3 = packet.data[10];
                    vdi2 = packet.data[11];
                    vdf2 = packet.data[12];
                    timer2 = packet.data[13] << 8 | packet.data[14];
                    timer4 = packet.data[15];

                    for (i = 0; i < TOT_OUTPUTS; i++)
                        buffer[i] = packet.data[TOT_OUTPUTS-1-i];
                    
                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer) ||
                            check_activity_bitmap_dac(map)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //ESP_LOGI(MAIN_TAG, "Not possible, resource is already allocated");
                        //send_nack(packet.sender, ADD, packet.command.cmd, IMPOSSIBLE);
                        //break;
                    }

                    launch_scaled_dac_timer_task(timer1, timer2, -1, numdac, vdi1, 
                            vdf1, vdi2, vdf2, timer3, timer4, 
                            0, -1, 0, buffer, updatequeue);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTn_SCALED_DAC_ALARM:
                    CHECK_LEN(packet, 13);
                    CHECK_RESET(packet);
                    num = packet.data[0];
                    timer1 = packet.data[1] << 8 | packet.data[2];
                    numdac = packet.data[3];
                    vdi1 = packet.data[4];
                    vdf1 = packet.data[5];
                    timer3 = packet.data[6];
                    vdi2 = packet.data[7];
                    vdf2 = packet.data[8];
                    timer2 = packet.data[9] << 8 | packet.data[10];
                    timer4 = packet.data[11];
                    value = packet.data[12];
                    memset(buffer, 0, TOT_OUTPUTS);
                    set_single_bitmap((int)num, buffer);
                    map = 1 << numdac;
                    
                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer) ||
                            check_activity_bitmap_dac(map)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //ESP_LOGI(MAIN_TAG, "Not possible, resource is already allocated");
                        //send_nack(packet.sender, ADD, packet.command.cmd, IMPOSSIBLE);
                        //break;
                    }

                    launch_scaled_dac_timer_task(timer1, timer2, value, numdac, vdi1, 
                            vdf1, vdi2, vdf2, timer3, timer4,
                            0, -1, 0, buffer, updatequeue);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTn_SCALED_DAC_420MA:
                    CHECK_LEN(packet, 14);
                    CHECK_RESET(packet);
                    num = packet.data[0];
                    timer1 = packet.data[1] << 8 | packet.data[2];
                    numdac = packet.data[3];
                    vdi1 = packet.data[4];
                    vdf1 = packet.data[5];
                    timer3 = packet.data[6];
                    vdi2 = packet.data[7];
                    vdf2 = packet.data[8];
                    timer2 = packet.data[9] << 8 | packet.data[10];
                    timer4 = packet.data[11];
                    arg1 = packet.data[12];
                    arg2 = packet.data[13];
                    memset(buffer, 0, TOT_OUTPUTS);
                    set_single_bitmap((int)num, buffer);
                    map = 1 << numdac;
                    
                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer) ||
                            check_activity_bitmap_dac(map)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //ESP_LOGI(MAIN_TAG, "Not possible, resource is already allocated");
                        //send_nack(packet.sender, ADD, packet.command.cmd, IMPOSSIBLE);
                        //break;
                    }

                    get_420ma_state(tmpbuffer);
                    launch_scaled_dac_timer_task(timer1, timer2, -1, numdac, vdi1, 
                            vdf1, vdi2, vdf2, timer3, timer4,
                            arg2, arg1, tmpbuffer[arg1] > arg2, buffer, updatequeue);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;

                case SET_OUTPUTn_SCALED_DAC_420MA_ALARM:
                    CHECK_LEN(packet, 15);
                    CHECK_RESET(packet);
                    num = packet.data[0];
                    timer1 = packet.data[1] << 8 | packet.data[2];
                    numdac = packet.data[3];
                    vdi1 = packet.data[4];
                    vdf1 = packet.data[5];
                    timer3 = packet.data[6];
                    vdi2 = packet.data[7];
                    vdf2 = packet.data[8];
                    timer2 = packet.data[9] << 8 | packet.data[10];
                    timer4 = packet.data[11];
                    arg1 = packet.data[12];
                    arg2 = packet.data[13];
                    value = packet.data[14];
                    memset(buffer, 0, TOT_OUTPUTS);
                    set_single_bitmap((int)num, buffer);
                    map = 1 << numdac;
                    
                    //if (!check_bitmap_inclusion(buffer, config.outputs, TOT_OUTPUTS) ||
                            if (check_activity_bitmap_output(buffer) ||
                            check_activity_bitmap_dac(map)) {
                        ESP_LOGI(MAIN_TAG, "deleting previous task");
                        delete_tasks_by_output(buffer);
                        //ESP_LOGI(MAIN_TAG, "Not possible, resource is already allocated");
                        //send_nack(packet.sender, ADD, packet.command.cmd, IMPOSSIBLE);
                        //break;
                    }

                    get_420ma_state(tmpbuffer);
                    launch_scaled_dac_timer_task(timer1, timer2, value, numdac, vdi1, 
                            vdf1, vdi2, vdf2, timer3, timer4,
                            arg2, arg1, tmpbuffer[arg1] > arg2, buffer, updatequeue);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;


                case LCD_ONOFF:
                    CHECK_LEN(packet, 1);
                    if (packet.data[0]) {
                        //TODO: accendere LCD
                    } else {
                        //TODO: spegnere LCD
                    }
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;
                
                case SET_COLORE_LED_DEFAULT:
                    CHECK_LEN(packet, 2);
                    config.colore[0] = packet.data[0];
                    config.colore[1] = packet.data[1];
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    sequentialWrite_24XX16(EEPROM_ADD, ADD_REV_HIGH, ADD_REV_LOW, (uint8_t*)&config, sizeof(machine_configuration_t));
                    break;

                case SET_COLORE_LED:
                    CHECK_LEN(packet, 2);
                    set_color_state(packet.data);
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;

                case SET_PULSING_LED:
                    CHECK_LEN(packet, 2);
                    config.pulsing_led = packet.data[0];
                    config.f_pulsing_led = packet.data[1];
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    sequentialWrite_24XX16(EEPROM_ADD, ADD_REV_HIGH, ADD_REV_LOW, (uint8_t*)&config, sizeof(machine_configuration_t));
                    ESP_LOGI(MAIN_TAG, "pulsing led number %i", config.pulsing_led);
                    break;

                case GET_COLORE_LED:
                    CHECK_LEN(packet, 0);
                    get_color_state(buffer); 
                    send_packet(packet.sender, ADD, packet.command.cmd, buffer, 2);
                    break;
                
                case TAKE_OFFSET_PRESS:
                    //TODO: rifiuta se e' troppo presto
                    CHECK_LEN(packet, 0);
                    take_offset_press(&config);
                    sequentialWrite_24XX16(EEPROM_ADD, ADD_REV_HIGH, ADD_REV_LOW, (uint8_t*)&config, sizeof(machine_configuration_t));
                    send_ack(packet.sender, ADD, packet.command.cmd);
                    break;

                default:
                    send_nack(packet.sender, ADD, packet.command.cmd, UNKNOWN_CMD);
                    ESP_LOGI(MAIN_TAG, "unknown command: %04X", packet.command.cmd);
                    break;
            }

            //if (packet.data != NULL)
                //free(packet.data);
        }
    }
}

void app_main(void) {
    xTaskCreate(task_main, "MAIN", 8000, NULL, 2, NULL);

    vTaskDelete(NULL);
}