#include <string.h>
#include "packet_manager.h"
#include "esp_log.h"
#include "utility.h"

#define TAG "Packet manager"

static const uint8_t preamble[PREAMBLE_LEN] = {0x02, 0x01};

int find_preamble(uint8_t *data, int begin, int end) {
    int i;
    for (i = begin; i < end; i++) {
        if (memcmp(&data[i], preamble, PREAMBLE_LEN) == 0)
            return i;
    }

    return -1;
}

int new_packet_data(unsigned long lasttimestamp, unsigned long timestamp, uint8_t *buffer, int *bufindex,
                    uint8_t *newdata, int len, int max) {
    int     pstart;
    uint8_t receiving = find_preamble(buffer, 0, *bufindex) >= 0;

    if (receiving && timestamp - lasttimestamp > TIMEOUT_US) {
        receiving = 0;
        *bufindex = 0;
        ESP_LOGD(TAG, "timeout");
    }

    if (*bufindex + len >= max) {
        receiving = 0;
        *bufindex = 0;
        return -1;
    }

    if (!receiving) {
        if ((pstart = find_preamble(newdata, 0, len)) >= 0) {
            memcpy(&buffer[*bufindex], newdata, len);
            *bufindex += len;
            receiving = 1;
            return 0;
        } else {
            *bufindex = 0;
            return -2;
        }
    } else {
        memcpy(&buffer[*bufindex], newdata, len);
        *bufindex += len;
    }

    return 0;
}

//TODO: Organizza meglio i valori di ritorno e la costruzione del pacchetto
int elaborate_data(uint8_t *data, int *start, int len, packet_t *packet) {
    uint8_t crc;
    int     pstart;

    pstart = find_preamble(data, *start, len);

    if (pstart < 0)
        return -2;

    // Se viene trovato l'header e se ci sono almeno altri due byte
    // a seguire (per la lunghezza del pacchetto)
    // TODO: aggiungi anche l'header come lunghezza (HEADER_LEN)
    if (pstart + PREAMBLE_LEN + HEADER_LEN < len) {
        *start         = pstart;
        packet->lenght = data[*start + 2];

        // Il pacchetto non e' ancora completo
        if (*start + packet->lenght + PREAMBLE_LEN + 1 > len)
            return -1;

        // Il pacchetto non e' valido
        if (data[*start + PREAMBLE_LEN + 1] != 0) {
            *start += 4;
            return -3;
        }
        if (data[*start + PREAMBLE_LEN + 4] != 0) {
            *start += 7;
            return -4;
        }

        packet->crc = data[*start + packet->lenght + PREAMBLE_LEN + 1];
        ESP_LOGD(TAG, "Lunghezza pacchetto: %i", packet->lenght);
        //hexprint(&data[pstart + PREAMBLE_LEN + 1], packet->lenght, TAG);
        crc = simple_crc(&data[pstart + PREAMBLE_LEN + 1], packet->lenght);

        // CRC errato
        if (crc != packet->crc) {
            ESP_LOGD(TAG, "CRC calcolato: %02X ; CRC trovato: %02X", crc, packet->crc);
            *start += packet->lenght + 4;
            return -5;
        }

        packet->receiver = (uint16_t)data[pstart + PREAMBLE_LEN + 2] << 8 | (uint16_t)data[pstart + PREAMBLE_LEN + 3];
        packet->sender   = (uint16_t)data[pstart + PREAMBLE_LEN + 5] << 8 | (uint16_t)data[pstart + PREAMBLE_LEN + 6];
        packet->command.cmd =
            (uint16_t)data[pstart + PREAMBLE_LEN + 7] << 8 | (uint16_t)data[pstart + PREAMBLE_LEN + 8];
        if (packet->lenght <= 8) {
            //packet->data = NULL;
        } else {
            // Checking if there is a memory leak
            //packet->data = (uint8_t *)malloc(packet->lenght - 8);
            memcpy(packet->data, &data[*start + 11], packet->lenght - 8);
        }
        *start += packet->lenght + 4;
        return 0;
    } else {
        return -1;
    }
}


int validate_slave_state(uint8_t *buffer, int len, uint8_t add) {
    uint8_t crc;
    if (len < SLAVE_STATE_LEN)
        return -1;
    if (buffer[0] != 0xBE)
        return -2;
    if (buffer[1] != add)
        return -3;

    crc = custom_crc(&buffer[2], SLAVE_STATE_LEN-SLAVE_HEADER-1, 0xBE);
    if (crc != buffer[len-1])
        return -4;

    return 0;
}


int build_packet(uint16_t receiver, uint16_t sender, uint16_t command, uint8_t *data, int len, uint8_t *buffer) {
    memcpy(buffer, preamble, PREAMBLE_LEN);

    buffer[PREAMBLE_LEN + 0] = len + 8;
    buffer[PREAMBLE_LEN + 1] = 0x00;
    buffer[PREAMBLE_LEN + 2] = receiver >> 8;
    buffer[PREAMBLE_LEN + 3] = receiver;
    buffer[PREAMBLE_LEN + 4] = 0x00;
    buffer[PREAMBLE_LEN + 5] = sender >> 8;
    buffer[PREAMBLE_LEN + 6] = sender;
    buffer[PREAMBLE_LEN + 7] = command >> 8;
    buffer[PREAMBLE_LEN + 8] = command;

    memcpy(&buffer[PREAMBLE_LEN + 9], data, len);

    buffer[PREAMBLE_LEN + 8 + len + 1] = simple_crc(&buffer[PREAMBLE_LEN + 1], len + 8);

    return len + 12;
}
