#include "hardwareprofile.h"
#include "driver/uart.h"
#include "freertos/task.h"
#include "freertos/event_groups.h"
#include "freertos/timers.h"
#include "uart_primary.h"
#include "utility.h"
#include <string.h>
#include "state.h"
#include "packet_manager.h"

#define TAG "485"

void init_485_generic() {
    uart_config_t uart_config = {
        .baud_rate = 256000,
        .data_bits = UART_DATA_8_BITS,
        .parity    = UART_PARITY_DISABLE,
        .stop_bits = UART_STOP_BITS_1,
        .flow_ctrl = UART_HW_FLOWCTRL_DISABLE,
    };
    // Configure UART parameters
    ESP_ERROR_CHECK(uart_param_config(UART_NUM_2, &uart_config));

    uart_set_pin(UART_NUM_2, GENERIC485_TX, GENERIC485_RX, -1, -1);
    gpio_set_direction(GENERIC485_DIR, GPIO_MODE_OUTPUT);
    gpio_set_pull_mode(GENERIC485_DIR, GPIO_FLOATING);
    gpio_intr_disable(GENERIC485_DIR);

    uart_driver_install(UART_NUM_2, 2048, 2048, 10, NULL, 0);
}

void init_485_display() {
    uart_config_t uart_config = {
        .baud_rate = 115200,
        .data_bits = UART_DATA_8_BITS,
        .parity    = UART_PARITY_DISABLE,
        .stop_bits = UART_STOP_BITS_1,
        .flow_ctrl = UART_HW_FLOWCTRL_DISABLE,
    };
    // Configure UART parameters
    ESP_ERROR_CHECK(uart_param_config(UART_NUM_1, &uart_config));

    uart_set_pin(UART_NUM_1, DISPLAY485_TX, DISPLAY485_RX, -1, -1);
    gpio_set_direction(DISPLAY485_DIR, GPIO_MODE_OUTPUT);
    gpio_set_pull_mode(DISPLAY485_DIR, GPIO_FLOATING);
    gpio_intr_disable(DISPLAY485_DIR);

    uart_driver_install(UART_NUM_1, 2048, 2048, 10, NULL, 0);
}

void send485(uint8_t *buffer, int len) {
    gpio_set_level(GENERIC485_DIR, 1);
    // vTaskDelay(10 / portTICK_PERIOD_MS);
    uart_flush_input(UART_NUM_2);
    uart_write_bytes(UART_NUM_2, (char *)buffer, len);
    uart_wait_tx_done(UART_NUM_2, portMAX_DELAY);
    // Aspetta alcuni microsecondi per essere sicuro che la trasmissione sia terminata
    // ets_delay_us(100);
    gpio_set_level(GENERIC485_DIR, 0);
}

int receive485(uint8_t *buffer, int length, int wait) {
    size_t received;
    int    nulls = 0;
    gpio_set_level(GENERIC485_DIR, 0);
    // vTaskDelay(wait / portTICK_PERIOD_MS);

    // uart_get_buffered_data_len(UART_NUM_2, &received);
    received = uart_read_bytes(UART_NUM_2, (uint8_t *)buffer, length, wait / portTICK_PERIOD_MS);

    for (nulls = 0; nulls < received; nulls++)
        if (buffer[nulls] != 0)
            break;

    received += uart_read_bytes(UART_NUM_2, (uint8_t *)&buffer[received], nulls, 0);     // wait / portTICK_PERIOD_MS);
    shift_array_left(buffer, received, nulls);
    received -= nulls;

    return received;
}

void send485_state(uint8_t add, uint8_t *state, int len) {
    uint8_t buffer[32];
    buffer[0] = 0xBE;
    buffer[1] = add;
    buffer[2] = 0x01;
    memcpy(&buffer[3], state, len);
    buffer[3 + len] = custom_crc(state, len, 0xBE);
    // hexprint(buffer, len+4, TAG);
    send485(buffer, len + 4);
}

void task_slave_communication(void *args) {
    int                      slave, len, received, tmp;
    uint8_t                  bitmap[32];
    uint8_t                  buffer485[128];
    void **                  params     = (void **)args;
    QueueHandle_t *          queue      = (QueueHandle_t *)params[0];
    machine_configuration_t *config     = (machine_configuration_t *)params[1];
    EventGroupHandle_t *     eventgroup = (EventGroupHandle_t *)params[2];
    // Aggiorna gli slave
    init_485_generic();

    for (;;) {
        for (slave = 0; slave < NUM_SLAVES; slave++) {
            // FIXME: mettiti d'accordo con michele su come gestire la configurazione, per ora vuole che lo slave 1 sia
            // sempre attivo
            if (slave == 0 || is_slave_configured(config, slave)) {
                get_output_state(bitmap);
                turn_on_led(*config, bitmap);
                set_output_state(bitmap);
                len = pack_slave_state(buffer485, config->f_pulsing_led ? config->pulsing_led : 0, slave);
                send485_state(slave, buffer485, len);
                memset(buffer485, 0, SLAVE_STATE_LEN);
                received = receive485(buffer485, SLAVE_STATE_LEN, 4);
                // hexprint(buffer485, received, TAG);
                if (received == 0) {
                    // TODO: verifica che la configurazione escluda correttamente lo slave
                    ESP_LOGI(TAG, "no answer to %i bytes from %i", len, slave);
                } else {
                    if ((tmp = validate_slave_state(buffer485, received, slave)) == 0) {
                        // se lo stato e' cambiato notifica i processi
                        if (unpack_slave_state(&buffer485[2], slave) || 1)     // TODO: togli || 1
                            xEventGroupSetBits(*eventgroup, ADC_EVENT_BITS | INPUT_EVENT_BITS);
                    } else
                        ESP_LOGI(TAG, "Invalid response of %i bytes: %i", received, tmp);
                    xQueueSend(*queue, NULL, 0);
                }
            }
        }
        vTaskDelay(30 / portTICK_PERIOD_MS);
    }
    vTaskDelete(NULL);
}