#include "hardwareprofile.h"
#include "esp_log.h"
#include <string.h>
#include "driver/uart.h"
#include "uart_primary.h"
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "utility.h"
#include "esp_system.h"

#include "packet_manager.h"

#define TAG "Primary UART"

static uint8_t  uart0_buffer[PACKET_BUFFER_SIZE] = {0};
static int      data_index                       = 0;
static unsigned last_timestamp                   = 0;

static QueueHandle_t uart0_queue;

void uart_primary_task(void *args) {
    uart_event_t   event;
    uint8_t        tmpbuffer[UART_PRIMARY_BUFFER_SIZE];
    int            res, start;
    packet_t       packet;
    QueueHandle_t *eventqueue = (QueueHandle_t *)args;
    for (;;) {
        // Waiting for UART event.
        if (xQueueReceive(uart0_queue, (void *)&event, (portTickType)portMAX_DELAY)) {
            switch (event.type) {
                case UART_DATA:
                    ESP_LOGD(TAG, "[UART DATA]: %d", event.size);
                    uart_read_bytes(UART_NUM_0, tmpbuffer, event.size, 1);
                    //if (esp_timer_get_time() - lasttimestamp > TIMEOUT_US)
                        //send_nack(0x00, 0x00, 0x0000, WRONG_CRC); //WRONG CRC

                    if ((res = new_packet_data(last_timestamp, esp_timer_get_time(), uart0_buffer, &data_index,
                                               tmpbuffer, event.size, PACKET_BUFFER_SIZE)) == 0) {
                        start = 0;
                        while ((res = elaborate_data(uart0_buffer, &start, data_index, &packet)) == 0) {
                            if (res == 0) {
                                if (packet.command.cmd == RESTART_BOARD) {
                                    send_ack(packet.sender, 0x0202, packet.command.cmd);
                                    esp_restart();
                                }
                                else if (xQueueSend(*eventqueue, &packet, 10 / portTICK_PERIOD_MS) != pdTRUE) {
                                    send_nack(packet.sender, 0x0202, packet.command.cmd, QUEUE_FULL);
                                }
                            }
                            else if (res == -5) {
                                send_nack(0x00, 0x00, 0x0000, WRONG_CRC); //WRONG CRC
                            }
                        }
                        if (res != -1 || data_index >= PACKET_BUFFER_SIZE)
                            data_index = 0;

                    } else {
                        ESP_LOGD(TAG, "Error result: %i", res);
                        data_index = 0;
                    }

                    last_timestamp = esp_timer_get_time();
                    break;
                // Event of HW FIFO overflow detected
                case UART_FIFO_OVF:
                    ESP_LOGD(TAG, "hw fifo overflow");
                    // If fifo overflow happened, you should consider adding flow control for your application.
                    // The ISR has already reset the rx FIFO,
                    // As an example, we directly flush the rx buffer here in order to read more data.
                    uart_flush_input(UART_NUM_0);
                    xQueueReset(uart0_queue);
                    data_index = 0;
                    break;
                // Event of UART ring buffer full
                case UART_BUFFER_FULL:
                    ESP_LOGD(TAG, "ring buffer full");
                    // If buffer full happened, you should consider encreasing your buffer size
                    // As an example, we directly flush the rx buffer here in order to read more data.
                    uart_flush_input(UART_NUM_0);
                    xQueueReset(uart0_queue);
                    data_index = 0;
                    break;
                // Event of UART RX break detected
                case UART_BREAK:
                    ESP_LOGD(TAG, "uart rx break");
                    break;
                // Event of UART parity check error
                case UART_PARITY_ERR:
                    ESP_LOGD(TAG, "uart parity error");
                    break;
                // Event of UART frame error
                case UART_FRAME_ERR:
                    ESP_LOGD(TAG, "uart frame error");
                    break;
                default:
                    ESP_LOGD(TAG, "uart event type: %d", event.type);
                    break;
            }
        }
    }
    vTaskDelete(NULL);
}

void init_uart_primary(QueueHandle_t *queue) {
    uart_config_t uart_config = {.baud_rate = 115200,
                                 .data_bits = UART_DATA_8_BITS,
                                 .parity    = UART_PARITY_DISABLE,
                                 .stop_bits = UART_STOP_BITS_1,
                                 .flow_ctrl = UART_HW_FLOWCTRL_DISABLE};
    uart_param_config(UART_NUM_0, &uart_config);

    // Install UART driver, and get the queue.
    uart_driver_install(UART_NUM_0, UART_PRIMARY_BUFFER_SIZE, UART_PRIMARY_BUFFER_SIZE, 20, &uart0_queue, 0);

    xTaskCreate(uart_primary_task, "uart primary", 4096, (void *)queue, 6, NULL);
}


void send_packet(uint16_t receiver, uint16_t sender, uint16_t command, uint8_t *data, int len) {
    uint8_t buffer[255];
    int     total;
    if (CONFIG_LOG_DEFAULT_LEVEL == ESP_LOG_INFO) {
        ESP_LOGD(TAG, "Packet sent:");
        total = build_packet(receiver, sender, command, data, len, buffer);
        //hexprint(buffer, total, TAG);
    } else {
        total = build_packet(receiver, sender, command, data, len, buffer);
        uart_write_bytes(UART_NUM_0, (char *)buffer, total);
    }
}

void send_ack(uint16_t receiver, uint16_t sender, uint16_t command) {
    send_packet(receiver, sender, command, NULL, 0);
}

void send_nack(uint16_t receiver, uint16_t sender, uint16_t command, uint8_t errorcode) {
    send_packet(receiver, sender, COMMAND_NACK(command), &errorcode, 1);
}
